# Copyright (C) 2011 by Canonical Ltd.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from ijson.lib import yajl
from ijson.parse import *

# from ijson.parse
_callback_data = [
    # Mapping of JSON parser events to callback C types and value converters.
    # Used to define the Callbacks structure and actual callback functions
    # inside the parse function.
    ('null', C_EMPTY, lambda: None),
    ('boolean', C_INT, lambda v: bool(v)),
    # "integer" and "double" aren't actually yielded by yajl since "number"
    # takes precedence if defined
    ('integer', C_LONG, lambda v, l: int(string_at(v, l))),
    ('double', C_DOUBLE, lambda v, l: float(string_at(v, l))),
    ('number', C_STR, lambda v, l: number(string_at(v, l))),
    ('string', C_STR, lambda v, l: string_at(v, l).decode('utf-8')),
    ('start_map', C_EMPTY, lambda: None),
    ('map_key', C_STR, lambda v, l: string_at(v, l)),
    ('end_map', C_EMPTY, lambda: None),
    ('start_array', C_EMPTY, lambda: None),
    ('end_array', C_EMPTY, lambda: None),
]


class Stream(object):
    def __init__(self):
        self.builder = None
        self.level = 0
        self.callbacks = Callbacks(*[self._callback(*data)
                                         for data in _callback_data])
        self.config = Config(False, True)
        self.handle = yajl.yajl_alloc(byref(self.callbacks), byref(self.config),
                                      None, None)
        self.objects = []

    def close(self):
        if self.handle:
            yajl.yajl_free(self.handle)
            self.handle = None

    def _callback(self, event, func_type, func):
        def c_callback(context, *args):
            if self.builder == None and event == 'start_map':
                self.builder = ObjectBuilder()
            self.builder.event(event, func(*args))

            if event == 'start_map':
                self.level += 1
            elif event == 'end_map':
                self.level -= 1

            if self.level == 0:
                self.objects.append(self.builder.value)
                self.builder = None

            return 1

        return func_type(c_callback)

    def _restart(self):
        yajl.yajl_parse_complete(self.handle)
        yajl.yajl_free(self.handle)
        self.handle = yajl.yajl_alloc(byref(self.callbacks), byref(self.config),
                                      None, None)

    def addData(self, data):
        while True:
            result = yajl.yajl_parse(self.handle, data, len(data))

            if result == YAJL_ERROR:
                perror = yajl.yajl_get_error(self.handle, 1, data, len(data))
                error = cast(perror, c_char_p).value
                yajl.yajl_free_error(self.handle, perror)

                raise JSONError(error)

            consumed = yajl.yajl_get_bytes_consumed(self.handle)
            if result == YAJL_OK:
                self._restart()

            remaining = len(data) - consumed

            if remaining == 0:
                return
            data = data[consumed:]

    def hasObject(self):
        return len(self.objects)

    def nextObject(self):
        return self.objects.pop(0)
